@@ -21,6 +21,7 @@ agent_local_config_iv = t.Dict({
         tx.AliasedKey(['backend', 'mode']): tx.Enum(AgentBackend),
         t.Key('rpc-listen-addr', default=('', 6001)):
             tx.HostPortPair(allow_blank_host=True),
+        t.Key('agent-sock-port', default=6007): t.Int[1024:65535],
         t.Key('id', default=None): t.Null | t.String,
         t.Key('region', default=None): t.Null | t.String,
         t.Key('instance-type', default=None): t.Null | t.String,
@@ -55,6 +56,9 @@ agent_local_config_iv = t.Dict({
         t.Key('enabled', default=False): t.Bool,
         t.Key('skip-container-deletion', default=False): t.Bool,
         t.Key('log-stats', default=False): t.Bool,
+        t.Key('log-kernel-config', default=False): t.Bool,
+        t.Key('log-alloc-map', default=False): t.Bool,
+        t.Key('log-events', default=False): t.Bool,
         t.Key('log-heartbeats', default=False): t.Bool,
         t.Key('log-docker-events', default=False): t.Bool,
         t.Key('coredump', default=coredump_defaults): t.Dict({
@@ -69,6 +73,12 @@ agent_local_config_iv = t.Dict({
     }).allow_extra('*'),
 }).merge(config.etcd_config_iv).allow_extra('*')
 
+docker_extra_config_iv = t.Dict({
+    t.Key('container'): t.Dict({
+        t.Key('swarm-enabled', default=False): t.Bool,
+    }).allow_extra('*')
+}).allow_extra('*')
+
 default_container_logs_config = {
     'max-length': '10M',  # the maximum tail size
     'chunk-size': '64K',  # used when storing logs to Redis as a side-channel to the event bus
