@@ -18,6 +18,7 @@ from typing import (
 import aiohttp
 from aiodocker.docker import Docker, DockerContainer
 from aiodocker.exceptions import DockerError
+import async_timeout
 import psutil
 
 from ai.backend.common.logging import BraceStyleAdapter
@@ -116,7 +117,7 @@ class CPUPlugin(AbstractComputePlugin):
     async def list_devices(self) -> Collection[CPUDevice]:
         cores = await libnuma.get_available_cores()
         overcommit_factor = int(os.environ.get('BACKEND_CPU_OVERCOMMIT_FACTOR', '1'))
-        assert 1 <= overcommit_factor <= 4
+        assert 1 <= overcommit_factor <= 10
         return [
             CPUDevice(
                 device_id=DeviceId(str(core_idx)),
@@ -188,7 +189,11 @@ class CPUPlugin(AbstractComputePlugin):
 
         async def api_impl(container_id):
             container = DockerContainer(ctx.agent.docker, id=container_id)
-            ret = await fetch_api_stats(container)
+            try:
+                async with async_timeout.timeout(2.0):
+                    ret = await fetch_api_stats(container)
+            except asyncio.TimeoutError:
+                return None
             if ret is None:
                 return None
             cpu_used = nmget(ret, 'cpu_stats.cpu_usage.total_usage', 0) / 1e6
@@ -324,11 +329,12 @@ class MemoryPlugin(AbstractComputePlugin):
     async def list_devices(self) -> Collection[MemoryDevice]:
         # TODO: support NUMA?
         memory_size = psutil.virtual_memory().total
+        overcommit_factor = int(os.environ.get('BACKEND_MEM_OVERCOMMIT_FACTOR', '1'))
         return [MemoryDevice(
             device_id=DeviceId('root'),
             hw_location='root',
             numa_node=0,
-            memory_size=memory_size,
+            memory_size=overcommit_factor * memory_size,
             processing_units=0,
         )]
 
@@ -458,7 +464,11 @@ class MemoryPlugin(AbstractComputePlugin):
 
         async def api_impl(container_id):
             container = DockerContainer(ctx.agent.docker, id=container_id)
-            ret = await fetch_api_stats(container)
+            try:
+                async with async_timeout.timeout(2.0):
+                    ret = await fetch_api_stats(container)
+            except asyncio.TimeoutError:
+                return None
             if ret is None:
                 return None
             mem_cur_bytes = nmget(ret, 'memory_stats.usage', 0)
