@@ -10,7 +10,6 @@ from pathlib import Path
 from pprint import pformat, pprint
 import signal
 import sys
-import traceback
 from typing import (
     Any,
     AsyncGenerator,
@@ -38,7 +37,6 @@ from trafaret.dataerror import DataError as TrafaretDataError
 
 from ai.backend.common import config, utils, identity, msgpack
 from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
-from ai.backend.common.events import AgentErrorEvent
 from ai.backend.common.logging import Logger, BraceStyleAdapter
 from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
 from ai.backend.common.types import (
@@ -112,10 +110,7 @@ def collect_error(meth: Callable) -> Callable:
         try:
             return await meth(self, *args, **kwargs)
         except Exception:
-            exc_type, exc, tb = sys.exc_info()
-            pretty_message = ''.join(traceback.format_exception_only(exc_type, exc)).strip()
-            pretty_tb = ''.join(traceback.format_tb(tb)).strip()
-            await self.agent.produce_event(AgentErrorEvent(pretty_message, pretty_tb))
+            await self.agent.produce_error_event()
             raise
     return _inner
 
@@ -183,6 +178,8 @@ class AgentRPCServer(aobject):
 
     async def __ainit__(self) -> None:
         # Start serving requests.
+        self._create_sema = asyncio.Semaphore(4)
+        self._destroy_sema = asyncio.Semaphore(4)
         await self.update_status('starting')
 
         if not self.skip_detect_manager:
@@ -317,27 +314,30 @@ class AgentRPCServer(aobject):
         raw_configs: Sequence[dict],
         raw_cluster_info: dict,
     ):
-        cluster_info = cast(ClusterInfo, raw_cluster_info)
-        session_id = SessionId(UUID(raw_session_id))
-        raw_results = []
-        coros = []
-        for raw_kernel_id, raw_config in zip(raw_kernel_ids, raw_configs):
-            log.info('rpc::create_kernel(k:{0}, img:{1})',
-                     raw_kernel_id, raw_config['image']['canonical'])
-            kernel_id = KernelId(UUID(raw_kernel_id))
-            kernel_config = cast(KernelCreationConfig, raw_config)
-            coros.append(self.agent.create_kernel(
-                creation_id,
-                session_id,
-                kernel_id,
-                kernel_config,
-                cluster_info,
-            ))
-        results = await asyncio.gather(*coros, return_exceptions=True)
+        async with self._create_sema:
+            cluster_info = cast(ClusterInfo, raw_cluster_info)
+            session_id = SessionId(UUID(raw_session_id))
+            raw_results = []
+            coros = []
+            for raw_kernel_id, raw_config in zip(raw_kernel_ids, raw_configs):
+                log.info('rpc::create_kernel(k:{0}, img:{1})',
+                        raw_kernel_id, raw_config['image']['canonical'])
+                kernel_id = KernelId(UUID(raw_kernel_id))
+                kernel_config = cast(KernelCreationConfig, raw_config)
+                coros.append(self.agent.create_kernel(
+                    creation_id,
+                    session_id,
+                    kernel_id,
+                    kernel_config,
+                    cluster_info,
+                ))
+            results = await asyncio.gather(*coros, return_exceptions=True)
         errors = [*filter(lambda item: isinstance(item, Exception), results)]
         if errors:
             # Raise up the first error.
-            raise errors[0]
+            if len(errors) == 1:
+                raise errors[0]
+            raise aiotools.MultiError("agent.create_kernels() failed", errors=errors)
         raw_results = [
             {
                 'id': str(result['id']),
@@ -363,16 +363,17 @@ class AgentRPCServer(aobject):
         reason: str = None,
         suppress_events: bool = False,
     ):
-        log.info('rpc::destroy_kernel(k:{0})', kernel_id)
-        done = asyncio.Event()
-        await self.agent.inject_container_lifecycle_event(
-            KernelId(UUID(kernel_id)),
-            LifecycleEvent.DESTROY,
-            reason or 'user-requested',
-            done_event=done,
-            suppress_events=suppress_events,
-        )
-        await done.wait()
+        async with self._destroy_sema:
+            log.info('rpc::destroy_kernel(k:{0})', kernel_id)
+            done = asyncio.Event()
+            await self.agent.inject_container_lifecycle_event(
+                KernelId(UUID(kernel_id)),
+                LifecycleEvent.DESTROY,
+                reason or 'user-requested',
+                done_event=done,
+                suppress_events=suppress_events,
+            )
+            await done.wait()
         return getattr(done, '_result', None)
 
     @rpc_function
